@@ -499,18 +499,18 @@ public:
     );
   }
 
-  void drawLink(pdf::link::Link *link, pdf::Catalog *catalog)
-  {
-     pdf::Renderer::drawLink(link, catalog); /* just call the inherited method */
-  }
-
-  void drawLink(pdf::link::Link *link, const std::string &border_color, pdf::Catalog *catalog)
+  void draw_link(pdf::link::Link *link, const std::string &border_color)
   {
     sexpr::GCLock gc_lock;
     if (!config.hyperlinks.extract)
       return;
     double x1, y1, x2, y2;
     pdf::link::Action *link_action = link->getAction();
+    if (link_action == NULL)
+    {
+      debug(1) << _("Warning: Unable to convert link without an action") << std::endl;
+      return;
+    }
     std::string uri;
     link->getRect(&x1, &y1, &x2, &y2);
     switch (link_action->getKind())
@@ -523,7 +523,7 @@ public:
       int page;
       try
       {
-        page = get_page_for_LinkGoTo(dynamic_cast<pdf::link::GoTo*>(link_action), catalog);
+        page = get_page_for_LinkGoTo(dynamic_cast<pdf::link::GoTo*>(link_action), this->catalog);
       }
       catch (NoLinkDestination &ex)
       {
@@ -1387,13 +1387,13 @@ static int xmain(int argc, char * const argv[])
         debug(0)++;
       }
       out1.reset(new pdf::Renderer(paper_color, config.monochrome));
-      out1->startDoc(doc->getXRef());
+      out1->start_doc(doc.get());
       outm.reset(new MutedRenderer(paper_color, config.monochrome, *page_files));
-      outm->startDoc(doc->getXRef());
+      outm->start_doc(doc.get());
       if (!config.monochrome)
       {
         outs.reset(new MutedRenderer(paper_color, config.monochrome, *page_files));
-        outs->startDoc(doc->getXRef());
+        outs->start_doc(doc.get());
       }
     }
     assert(doc.get() != NULL);
