@@ -64,54 +64,103 @@ std::string POSIXError::error_message(const std::string &context)
   /* Win32 systems tends to return POSIX-style error messages in English.
    * Translation is required.
    */
+  /* L10N: Windows error message for EPERM */
   N_("Operation not permitted");
+  /* L10N: Windows error message for ENOENT */
   N_("No such file or directory");
+  /* L10N: Windows error message for ESRCH */
   N_("No such process");
+  /* L10N: Windows error message for EINTR */
   N_("Interrupted function call");
+  /* L10N: Windows error message for EINTR */
   N_("Interrupted system call");
+  /* L10N: Windows error message for EIO */
   N_("Input/output error");
+  /* L10N: Windows error message for ENXIO */
   N_("No such device or address");
+  /* L10N: Windows error message for E2BIG */
   N_("Arg list too long");
+  /* L10N: Windows error message for E2BIG */
   N_("Argument list too long");
+  /* L10N: Windows error message for ENOEXEC */
   N_("Exec format error");
+  /* L10N: Windows error message for EBADF */
   N_("Bad file descriptor");
+  /* L10N: Windows error message for ECHILD */
   N_("No child processes");
+  /* L10N: Windows error message for EAGAIN */
   N_("Resource temporarily unavailable");
+  /* L10N: Windows error message for ENOMEM */
   N_("Not enough space");
+  /* L10N: Windows error message for ENOMEM */
   N_("Cannot allocate memory");
+  /* L10N: Windows error message for EACCES */
   N_("Permission denied");
+  /* L10N: Windows error message for EFAULT */
   N_("Bad address");
+  /* L10N: Windows error message for EBUSY */
   N_("Resource device");
+  /* L10N: Windows error message for EBUSY */
   N_("Device or resource busy");
+  /* L10N: Windows error message for EEXIST */
   N_("File exists");
+  /* L10N: Windows error message for EXDEV */
   N_("Improper link");
+  /* L10N: Windows error message for EXDEV */
   N_("Invalid cross-device link");
+  /* L10N: Windows error message for ENODEV */
   N_("No such device");
+  /* L10N: Windows error message for ENOTDIR */
   N_("Not a directory");
+  /* L10N: Windows error message for EISDIR */
   N_("Is a directory");
+  /* L10N: Windows error message for EINVAL */
   N_("Invalid argument");
+  /* L10N: Windows error message for ENFILE */
   N_("Too many open files in system");
+  /* L10N: Windows error message for EMFILE */
   N_("Too many open files");
+  /* L10N: Windows error message for ENOTTY */
   N_("Inappropriate I/O control operation");
+  /* L10N: Windows error message for ENOTTY */
   N_("Inappropriate ioctl for device");
+  /* L10N: Windows error message for EFBIG */
   N_("File too large");
+  /* L10N: Windows error message for ENOSPC */
   N_("No space left on device");
+  /* L10N: Windows error message for ESEEK */
   N_("Invalid seek");
+  /* L10N: Windows error message for ESEEK */
   N_("Illegal seek");
+  /* L10N: Windows error message for EROFS */
   N_("Read-only file system");
+  /* L10N: Windows error message for EMLINK */
   N_("Too many links");
+  /* L10N: Windows error message for EPIPE */
   N_("Broken pipe");
+  /* L10N: Windows error message for EDOM */
   N_("Domain error");
+  /* L10N: Windows error message for EDOM */
   N_("Numerical argument out of domain");
+  /* L10N: Windows error message for ERANGE */
   N_("Result too large");
+  /* L10N: Windows error message for ERANGE */
   N_("Numerical result out of range");
+  /* L10N: Windows error message for EDEADLK */
   N_("Resource deadlock avoided");
+  /* L10N: Windows error message for ENAMETOOLONG */
   N_("Filename too long");
+  /* L10N: Windows error message for ENAMETOOLONG */
   N_("File name too long");
+  /* L10N: Windows error message for ENOLOCK */
   N_("No locks available");
+  /* L10N: Windows error message for ENOSYS */
   N_("Function not implemented");
+  /* L10N: Windows error message for ENOTEMPTY */
   N_("Directory not empty");
+  /* L10N: Windows error message for EILSEQ */
   N_("Illegal byte sequence");
+  /* L10N: Windows error message for EILSEQ */
   N_("Invalid or incomplete multibyte or wide character");
   message = _(strerror(errno));
 #else
@@ -349,14 +398,23 @@ void Command::call(std::ostream *my_stdout, bool quiet)
   status = xsystem.rdbuf()->status();
   if (status != 0)
   {
-    std::ostringstream message;
-    message
-      << "system(\""
-      << this->command
-      << " ...\") failed";
+    std::string message;
     if (WIFEXITED(status))
-      message << " with exit code " << WEXITSTATUS(status);
-    throw CommandFailed(message.str());
+    {
+      message = string_printf(
+        _("External command \"%s ...\" failed with exit code %u"),
+        this->command.c_str(),
+        static_cast<unsigned int>(WEXITSTATUS(status))
+      );
+    }
+    else
+    {
+      message = string_printf(
+        _("External command \"%s ...\" failed"),
+        this->command.c_str()
+      );
+    }
+    throw CommandFailed(message);
   }
 }
 
@@ -366,8 +424,10 @@ void Command::call(std::ostream *my_stdout, bool quiet)
 {
   int status = 0;
   unsigned long rc;
+  PROCESS_INFORMATION process_info;
   HANDLE read_end, write_end, error_handle;
   SECURITY_ATTRIBUTES security_attributes;
+  memset(&process_info, 0, sizeof process_info);
   security_attributes.nLength = sizeof (SECURITY_ATTRIBUTES);
   security_attributes.lpSecurityDescriptor = NULL;
   security_attributes.bInheritHandle = true;
@@ -411,8 +471,6 @@ void Command::call(std::ostream *my_stdout, bool quiet)
   }
   {
     const std::string &command_line = argv_to_command_line(this->argv);
-    PROCESS_INFORMATION process_info;
-    memset(&process_info, 0, sizeof process_info);
     STARTUPINFO startup_info;
     memset(&startup_info, 0, sizeof startup_info);
     startup_info.cb = sizeof startup_info;
@@ -435,14 +493,13 @@ void Command::call(std::ostream *my_stdout, bool quiet)
       status = -1;
     else
     {
-      CloseHandle(process_info.hProcess); /* ignore errors */
-      CloseHandle(process_info.hThread); /* ignore errors */
       if (error_handle != INVALID_HANDLE_VALUE)
         CloseHandle(error_handle); /* ignore errors */
     }
   }
   if (status == 0)
   {
+    unsigned long exit_code;
     CloseHandle(write_end); /* ignore errors */
     while (true)
     {
@@ -458,15 +515,31 @@ void Command::call(std::ostream *my_stdout, bool quiet)
         my_stdout->write(buffer, nbytes);
     }
     CloseHandle(read_end); /* ignore errors */
+    rc = WaitForSingleObject(process_info.hProcess, INFINITE);
+    if (rc == WAIT_FAILED)
+      throw_win32_error("WaitForSingleObject");
+    rc = GetExitCodeProcess(process_info.hProcess, &exit_code);
+    CloseHandle(process_info.hProcess); /* ignore errors */
+    CloseHandle(process_info.hThread); /* ignore errors */
+    if (rc == 0)
+      status = -1;
+    else if (exit_code != 0)
+    {
+      std::string message = string_printf(
+        _("External command \"%s ...\" failed with exit code %u"),
+        this->command.c_str(),
+        static_cast<unsigned int>(exit_code)
+      );
+      throw CommandFailed(message);
+    }
   }
   if (status < 0)
   {
-    std::ostringstream message;
-    message
-      << "system(\""
-      << this->command
-      << " ...\") failed";
-    throw Win32Error(message.str());
+    std::string message = string_printf(
+      _("External command \"%s ...\" failed"),
+      this->command.c_str()
+    );
+    throw Win32Error(message);
   }
 }
 
@@ -496,18 +569,37 @@ void Command::call(std::ostream *my_stdout, bool quiet)
       if (my_stdout != NULL)
         my_stdout->write(buffer, nbytes);
     }
-    status = status || pclose(file);
+    int wait_status = pclose(file);
+    if (wait_status == -1)
+      throw_posix_error("pclose");
+    else if (status == 0)
+    {
+      if (WIFEXITED(wait_status))
+      {
+        unsigned int exit_code = WEXITSTATUS(wait_status);
+        if (exit_code != 0)
+        {
+          std::string message = string_printf(
+            _("External command \"%s ...\" failed with exit code %u"),
+            this->command.c_str(),
+            exit_code
+          );
+          throw CommandFailed(message);
+        }
+      }
+      else
+        status = -1;
+    }
   }
   else
     status = -1;
   if (status < 0)
   {
-    std::ostringstream message;
-    message
-      << "system(\""
-      << this->command
-      << " ...\") failed";
-    throw_posix_error(message.str());
+    std::string message = string_printf(
+      _("External command \"%s ...\" failed"),
+      this->command.c_str()
+    );
+    throw_posix_error(message);
   }
 }
 
@@ -544,7 +636,9 @@ std::string Command::filter(const std::string &command_line, const std::string s
   int status = 0;
   unsigned long rc;
   HANDLE stdin_read, stdin_write, stdout_read, stdout_write, error_handle;
+  PROCESS_INFORMATION process_info;
   SECURITY_ATTRIBUTES security_attributes;
+  memset(&process_info, 0, sizeof process_info);
   security_attributes.nLength = sizeof (SECURITY_ATTRIBUTES);
   security_attributes.lpSecurityDescriptor = NULL;
   security_attributes.bInheritHandle = true;
@@ -578,8 +672,6 @@ std::string Command::filter(const std::string &command_line, const std::string s
       throw_win32_error("DuplicateHandle");
   }
   {
-    PROCESS_INFORMATION process_info;
-    memset(&process_info, 0, sizeof process_info);
     STARTUPINFO startup_info;
     memset(&startup_info, 0, sizeof startup_info);
     startup_info.cb = sizeof startup_info;
@@ -603,14 +695,13 @@ std::string Command::filter(const std::string &command_line, const std::string s
       status = -1;
     else
     {
-      CloseHandle(process_info.hProcess); /* ignore errors */
-      CloseHandle(process_info.hThread); /* ignore errors */
       if (error_handle != INVALID_HANDLE_VALUE)
         CloseHandle(error_handle); /* ignore errors */
     }
   }
   if (status == 0)
   {
+    unsigned long exit_code;
     std::ostringstream stream;
     CloseHandle(stdin_read); /* ignore errors */
     CloseHandle(stdout_write); /* ignore errors */
@@ -635,16 +726,32 @@ std::string Command::filter(const std::string &command_line, const std::string s
     if (rc == WAIT_FAILED)
       throw_win32_error("WaitForSingleObject");
     CloseHandle(thread_handle); /* ignore errors */
+    rc = WaitForSingleObject(process_info.hProcess, INFINITE);
+    if (rc == WAIT_FAILED)
+      throw_win32_error("WaitForSingleObject");
+    rc = GetExitCodeProcess(process_info.hProcess, &exit_code);
+    CloseHandle(process_info.hProcess); /* ignore errors */
+    CloseHandle(process_info.hThread); /* ignore errors */
+    if (rc == 0)
+      status = -1;
+    else if (exit_code != 0)
+    {
+      std::string message = string_printf(
+        _("External command \"%s\" failed with exit code %u"),
+        command_line.c_str(),
+        static_cast<unsigned int>(exit_code)
+      );
+      throw CommandFailed(message);
+    }
     return stream.str();
   }
   if (status < 0)
   {
-    std::ostringstream message;
-    message
-      << "system(\""
-      << command_line
-      << "\") failed";
-    throw Win32Error(message.str());
+    std::string message = string_printf(
+      _("External command \"%s\" failed"),
+      command_line.c_str()
+    );
+    throw Win32Error(message);
   }
   return string; /* Should not really happen. */
 }
@@ -674,9 +781,13 @@ std::string Command::filter(const std::string &command_line, const std::string s
     if (fputs(string.c_str(), fp) == EOF)
       throw_posix_error("fputs");
     rc = pclose(fp);
-    if (rc != 0)
+    if (rc == -1)
       throw_posix_error("pclose");
-    exit(0);
+    else if (WIFEXITED(rc))
+      exit(WEXITSTATUS(rc));
+    else
+      exit(-1);
+    exit(rc);
   }
   else
   {
@@ -700,14 +811,23 @@ std::string Command::filter(const std::string &command_line, const std::string s
       throw_posix_error("wait");
     if (status != 0)
     {
-      std::ostringstream message;
-      message
-        << "system(\""
-        << command_line
-        << "\") failed";
+      std::string message;
       if (WIFEXITED(status))
-        message << " with exit code " << WEXITSTATUS(status);
-      throw CommandFailed(message.str());
+      {
+        message = string_printf(
+          _("External command \"%s\" failed with exit code %u"),
+          command_line.c_str(),
+          static_cast<unsigned int>(WEXITSTATUS(status))
+        );
+      }
+      else
+      {
+        message = string_printf(
+          _("External command \"%s\" failed"),
+          command_line.c_str()
+        );
+      }
+      throw CommandFailed(message);
     }
     return stream.str();
   }
